﻿using System;
using System.Collections;
using System.Collections.Generic;
using System.Linq;

using System.IO;
using System.Runtime.Serialization;
using System.Runtime.Serialization.Formatters.Binary;

using UnityEngine;
using UnityEngine.EventSystems;
using UnityEngine.UI;

public class GridLayout : MonoBehaviour
{
    public GridCell[] Grid  { get; set; }
    public int Width;
    public int Height;
    public bool MapEditor;
    public GameObject Panel;
    public Dropdown TileTypeDropDown;
    public Toggle CollidableCheckBox;
    public Text Text;
    public LayerMask LayerMask;
    public LayerMask UILayerMask;
    public Camera Camera;
    public Button SaveLevelButton;
    public GameObject[] TilePrefabs;
    public Button[] Loadbuttons;
    public InputField inputFieldEdytor;
    public InputField inputFieldConnection;


    //Map Editor Vars
    private GridCell CurrentGridCell = null;
    private GraphicRaycaster raycaster;
    private EventSystem eventSystem;
    private float timeCounter;

    private string UnityPath;
    public string defaultPath;

    //Drawing
    private List<GameObject> gameObjectsToDraw;
    public GameObject DrawableObject;
    private bool GridInicialized = false;


    #region Events

    public void Start()
    {
        UnityPath = Application.dataPath;
        defaultPath = UnityPath;
        inputFieldConnection.text = "HiLevel";

        timeCounter = Time.time * 1000;

        gameObjectsToDraw = new List<GameObject>();


        
        Grid = new GridCell[Width * Height];
        for (int x = 0; x < Width; x++)
        {
            for (int y = 0; y < Height; y++)
            {
                Grid[x + y * Width] = new GridCell(x, y);
            }
        }
        GridInicialized = true;
        

        Grid = GenerateGrid();
            

        Panel.SetActive(MapEditor);
        if(MapEditor)
        {
            TileTypeDropDown.options = new List<Dropdown.OptionData>();
            foreach (TileType t in (TileType[])Enum.GetValues(typeof(TileType)))
            {
                TileTypeDropDown.options.Add(new Dropdown.OptionData(t.ToString()));
            }
        }
        // UI Events

        foreach(Button b in Loadbuttons)
        {
            b.onClick.AddListener(delegate
            {
                OnLoadButtonClick(b);
            });

        }

        TileTypeDropDown.onValueChanged.AddListener(delegate {
            OnTileTypeDropDownValueChange(TileTypeDropDown);
        });
        SaveLevelButton.onClick.AddListener(delegate {
            OnSaveLevelButtonClick(SaveLevelButton);
        });

        raycaster = Panel.GetComponent<GraphicRaycaster>();
        eventSystem = GetComponent<EventSystem>();


        //Draw
        Draw();
    }

    void OnLoadButtonClick(Button sender)
    {
        string Name = string.Empty;
        if(MapEditor)
        {
            Name = inputFieldEdytor?.text;
        }
        else
        {
            Name = inputFieldConnection?.text;
        }
        if (LoadGridFromFile(defaultPath, out GridCell[] gridCell, Name))
        {
            Grid = gridCell;
            Draw();
        }
    }

    void OnSaveLevelButtonClick(Button sender)
    {
        string Name = string.Empty;
        if (MapEditor)
        {
            Name = inputFieldEdytor?.text;
        }
        SaveGridToFile(defaultPath, Name);
    }

    void OnTileTypeDropDownValueChange(Dropdown sender)
    {
        if(CurrentGridCell != null)
        {
            CurrentGridCell.TileType = (TileType)sender.value;
            SetGridCell(CurrentGridCell);
        }
    }

    public void Draw()
    {
        if(GridInicialized)
        {
            for(int i = 0; i < gameObjectsToDraw.Count; i++)
            {
                Destroy(gameObjectsToDraw[i]);
            }

            gameObjectsToDraw = new List<GameObject>();

            foreach(GridCell gc in Grid)
            {
                GameObject prefab;
                switch (gc.TileType)
                {
                    case TileType.Grass:
                        prefab = TilePrefabs[0];
                        break;
                    case TileType.Road:
                        prefab = TilePrefabs[1];
                        break;
                    case TileType.Scenery:
                        prefab = TilePrefabs[2];
                        break;
                    default:
                        prefab = TilePrefabs[3];
                        break;
                }
                if (GetWorldPosFromCellIndexes(gc.X, gc.Y, out Vector2 worldPos))
                    prefab.transform.position = new Vector3(worldPos.x, 0.0f, worldPos.y);
                gameObjectsToDraw.Add(Instantiate(prefab, prefab.transform.position, prefab.transform.rotation, DrawableObject.transform));
            }
        }
    }

    public void Update()
    {
        if(MapEditor)
        {
            if (Time.time * 1000 - timeCounter > 100)
            {
                Draw();
                timeCounter = Time.time * 1000;
            }
        }

        float horizontal = 0;
        float vertical = 0;
        if (Input.GetKey(KeyCode.I)) vertical += 1;
        if (Input.GetKey(KeyCode.K)) vertical -= 1;
        if (Input.GetKey(KeyCode.L)) horizontal += 1;
        if (Input.GetKey(KeyCode.J)) horizontal -= 1;

        float CameraSpeed = 5;
        Vector3 translation = new Vector3(horizontal * CameraSpeed, 0.0f, vertical * CameraSpeed);

        Camera.transform.Translate(translation, Space.World);
    }

    public void OnMouseDown()
    {
        if(MapEditor)
        {
            Vector3 mousePos = Input.mousePosition;
            if(mousePos != null)
            {
                Ray ray = Camera.ScreenPointToRay(mousePos);
                RaycastHit raycastHit;
                PointerEventData pointerEventData = new PointerEventData(eventSystem);
                pointerEventData.position = mousePos;
                List<RaycastResult> raycastResults = new List<RaycastResult>();
                raycaster.Raycast(pointerEventData, raycastResults);

                if (!(raycastResults.Count > 0))
                if (Physics.Raycast(ray, out raycastHit, 1000, LayerMask))
                {
                    Vector2 cellSize = GetGridCellSize();
                    Vector2 position = new Vector2(this.transform.position.x, this.transform.position.z);
                    Vector2 scale = GetScale() / 2;
                    Vector2 startingPoint = position - scale;
                    Vector2 point = new Vector2(raycastHit.point.x, raycastHit.point.z);
                    GridCell gridCell = null;

                    //Search for cell
                    if(GetCellIndexesFromWorldPos(point, out int x, out int y))
                        gridCell = GetGridCell(x, y);
                    
                    if(gridCell != null)
                    {
                        CurrentGridCell = gridCell;
                        //Update UI
                        Text.text = $"X: {gridCell.X}\nY: {gridCell.Y}";
                        TileTypeDropDown.value = (int)gridCell.TileType;
                    }
                }
            }
        }
    }

    #endregion

    #region Utils

    public GridCell[] GenerateGrid()
    {
        GridCell[] gridCells = new GridCell[Width * Height];
        for (int x = 0; x < Width; x++)
        {
            for (int y = 0; y < Height; y++)
            {
                gridCells[x + y * Width] = new GridCell(x, y);
            }
        }


        //TileType tileType = TileType.Grass;
        
        for(int x = 0; x < Width; x++)
        {
            for(int y = 0; y < Height; y++)
            {
                    GridCell gridCell = GetGridCell(x, y);

                int numberOfBuildings = GetNumberOfStraightMatchedCells(gridCell, TileType.Building, gridCells);
                int numberOfStraightRoads = GetNumberOfStraightMatchedCells(gridCell, TileType.Building, gridCells);
                int numberOfDiagonalRoads = GetNumberOfDiagonalMatchedCells(gridCell, TileType.Building, gridCells);

                //float BuildingsScore = numberOfBuildings * (-0.25f);
                //float DiagonalScore = numberOfDiagonalRoads * (2f);
                //float StraightScore = numberOfStraightRoads * ((numberOfStraightRoads <= 2) ? (-0.5f) : (0.50f));

                //if (num > BuildingsScore + DiagonalScore + StraightScore)
                //    tileType = TileType.Road;
                //else
                //    tileType = TileType.Scenery;
                
                //gridCells[x + y * Width].TileType = tileType;
            }
        }
        
        return gridCells;
    }


    #region Number Of Neighbours

    public int GetNumberOfStraightMatchedCells(GridCell gridCell, TileType testedTileType, GridCell[] gridCells)
    {
        int x = gridCell.X;
        int y = gridCell.Y;
        int X = Width;
        int Y = Height;
        int count = 0;
        //Adding top row
        if (y > 0)
        {
            
            if (x >= 0 && GetCellState(x, y - 1, testedTileType, gridCells)) count++;
            
        }
        //Adding middle row
        if (x > 0 && GetCellState(x - 1, y, testedTileType, gridCells)) count++;
        if (x < X - 1 && GetCellState(x + 1, y, testedTileType, gridCells)) count++;
        //Adding bottom row
        if (y < Y - 1)
        {
            
            if (x >= 0 && GetCellState(x, y + 1, testedTileType, gridCells)) count++;
            
        }
        return count;
    }

    public int GetNumberOfDiagonalMatchedCells(GridCell gridCell, TileType testedTileType, GridCell[] gridCells)
    {
        int x = gridCell.X;
        int y = gridCell.Y;
        int X = Width;
        int Y = Height;
        int count = 0;
        //Adding top row
        if (y > 0)
        {
            if (x > 0 && GetCellState(x - 1, y - 1, testedTileType, gridCells)) count++;
            
            if (x < X - 1 && GetCellState(x + 1, y - 1, testedTileType, gridCells)) count++;
        }
        //Adding middle row
        
        //Adding bottom row
        if (y < Y - 1)
        {
            if (x > 0 && GetCellState(x - 1, y + 1, testedTileType, gridCells)) count++;
            
            if (x < X - 1 && GetCellState(x + 1, y + 1, testedTileType, gridCells)) count++;
        }
        return count;
    }

    public int GetNumberOfMatchedCells(GridCell gridCell, TileType testedTileType, GridCell[] gridCells)
    {
        int x = gridCell.X;
        int y = gridCell.Y;
        int X = Width;
        int Y = Height;
        int count = 0;
        //Adding top row
        if (y > 0)
        {
            if (x > 0 && GetCellState(x - 1, y - 1, testedTileType, gridCells)) count++;
            if (x >= 0 && GetCellState(x, y - 1, testedTileType, gridCells)) count++;
            if (x < X - 1 && GetCellState(x + 1, y - 1, testedTileType, gridCells)) count++;
        }
        //Adding middle row
        if (x > 0 && GetCellState(x - 1, y, testedTileType, gridCells)) count++;
        if (x < X - 1 && GetCellState(x + 1, y, testedTileType, gridCells)) count++;
        //Adding bottom row
        if (y < Y - 1)
        {
            if (x > 0 && GetCellState(x - 1, y + 1, testedTileType, gridCells)) count++;
            if (x >= 0 && GetCellState(x, y + 1, testedTileType, gridCells)) count++;
            if (x < X - 1 && GetCellState(x + 1, y + 1, testedTileType, gridCells)) count++;
        }
        return count;
    }

    public bool GetCellState(int x, int y, TileType testedTileType, GridCell[] gridCells)
    {
        return (gridCells[x + y * Width].TileType == testedTileType);
    }

    #endregion

    #region Save Load

    public bool SaveGridToFile(string path, string name = "Level1" ,FileMode fileMode = FileMode.OpenOrCreate)
    {
        try
        {
            Stream stream = File.Open(path + @"\" + name + ".lvl", fileMode);
            BinaryFormatter bf = new BinaryFormatter();

            bf.Serialize(stream, Grid);

            stream.Close();
        }
        catch (Exception e)
        {
            print(e.Message);
            print(e.StackTrace);
            return false;
        }

        return true;
    }

    public bool LoadGridFromFile(string path, out GridCell[] gridCells, string name = "Level1")
    {
        gridCells = null;
        try
        {
            Stream stream = File.Open(path + @"\" + name + ".lvl", FileMode.Open);

            BinaryFormatter bf = new BinaryFormatter();

            gridCells = (GridCell[])bf.Deserialize(stream);

            stream.Close();
        } 
        catch (Exception e)
        {
            print(e.Message);
            print(e.StackTrace);
            return false;
        }

        return true;
    }

    #endregion

    public GridCell GetGridCell(int x, int y) 
    {
        return Grid[x + y * Width];
    }

    public bool SetGridCell(GridCell gridCell)
    {
        try
        {
            Grid[gridCell.X + gridCell.Y * Width] = gridCell;
        }
        catch
        {
            print($"SetGridCell: {gridCell} was not set");
            return false;
        }
        return true;
    }

    public Vector2 GetGridCellSize()
    {
        return new Vector2(this.transform.localScale.x * 10 / Width, this.transform.localScale.z * 10 / Height);
    }

    /// <summary>
    /// Returns 10 times the plane scale
    /// </summary>
    /// <returns></returns>
    public Vector2 GetScale()
    {
        return new Vector2(this.transform.localScale.x * 10, this.transform.localScale.z * 10);
    }

    public bool GetCellIndexesFromWorldPos(Vector2 WorldPos, out int X, out int Y)
    {
        X = Y = 0;
        bool found = false;

        try
        {
            Vector2 cellSize = GetGridCellSize();
            Vector2 position = new Vector2(this.transform.position.x, this.transform.position.z);
            Vector2 scale = GetScale() / 2;
            Vector2 startingPoint = position - scale;
            Vector2 point = WorldPos;

            for (int x = 0; x < Width; x++)
            {
                for (int y = 0; y < Height; y++)
                {
                    if (point.x >= startingPoint.x + cellSize.x * x &&
                        point.y >= startingPoint.y + cellSize.y * y &&
                        point.x < startingPoint.x + cellSize.x * x + cellSize.x &&
                        point.y < startingPoint.y + cellSize.y * y + cellSize.y)
                    {
                        X = x;
                        Y = y;
                        found = true;
                    }
                    if (found) break;
                }
                if (found) break;
            }
        }
        catch (Exception e)
        {
            print(e.StackTrace);
            return false;
        }

        return found;
    }

    public bool GetWorldPosFromCellIndexes(int x, int y, out Vector2 worldPos)
    {
        worldPos = new Vector2();
        try
        {
            Vector2 cellSize = GetGridCellSize();
            Vector2 position = new Vector2(this.transform.position.x, this.transform.position.z);
            Vector2 scale = GetScale() / 2;
            Vector2 startingPoint = position - scale;

            float xPos = startingPoint.x + cellSize.x * x + cellSize.x / 2;
            float yPos = startingPoint.y + cellSize.y * y + cellSize.y / 2;

            worldPos = new Vector2(xPos, yPos);
        }
        catch ( Exception e)
        {
            print(e.StackTrace);
            return false;
        }

        return true;
    }

    #endregion

}

[Serializable()]
public class GridCell : ISerializable
{
    public TileType TileType { get; set; }
    public Effect Effect { get; set; }
    public bool Collidable
    {
        get
        {
            if (TileType == TileType.Building)
                return true;
            else return false;
        }
    }
    public int X { get; protected set; }
    public int Y { get; protected set; }

    public GridCell(int x, int y, TileType tileType = TileType.Grass, Effect effect = Effect.None)
    {
        TileType = tileType;
        float num = UnityEngine.Random.Range(0.0f, 1.0f);
        if (num > 0.75f)
            TileType = TileType.Scenery;
        if (num > 0.94f)
            TileType = TileType.Building;
        Effect = effect;
        X = x;
        Y = y;
    }

    public override string ToString()
    {
        return $"Grid Cell X: {X} Y: {Y}, Collidable: {Collidable}, TileType: {TileType}, Effect: {Effect}";
    }

    public void GetObjectData(SerializationInfo info, StreamingContext context)
    {
        info.AddValue("TileType", TileType);
        info.AddValue("Effect", Effect);
        info.AddValue("X", X);
        info.AddValue("Y", Y);
    }

    public GridCell(SerializationInfo info, StreamingContext context)
    {
        TileType = (TileType)info.GetValue("TileType", typeof(TileType));
        Effect = (Effect)info.GetValue("Effect", typeof(Effect));
        X = (int)info.GetValue("X", typeof(int));
        Y = (int)info.GetValue("Y", typeof(int));
    }
}

public enum TileType
{
    Grass,
    Road,
    Scenery,
    Building
}

public enum Effect
{
    None
}
